"""
KARP - Server
"""

import asyncio
import logging
import traceback
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Optional, Dict

from karp.utils import PendingRequest, Utils
from karp.request import Request
from karp.response import Response, InvalidResponseException

logging.basicConfig(level=logging.WARNING)


class InvalidRouteNameException(Exception):
    """
    Raised if route does not exist
    """

    pass


class Client:
    """
    Client
    """
    def __init__(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, _id: str
    ):
        self.reader = reader
        self.writer = writer
        self.id = _id

        self.requests: Dict[str, PendingRequest] = {}

    async def request(
        self, route: str, data: str, response: bool = True, timeout: int = None
    ):
        """

        :param route:
        :param data:
        :param response:
        :param timeout:
        :return:
        """
        req: Request = Request.create(route, data, response)
        self.writer.write(req.to_bytes())

        if response:
            self.requests[req.request_id] = PendingRequest()
        await self.writer.drain()

        if not response:
            return

        try:
            response = await self.requests[req.request_id].process(
                timeout=timeout
            )
        except TimeoutError:
            del self.requests[req.request_id]
            raise

        del self.requests[req.request_id]
        return response


class KARPServer:
    """
    KARPServer
    """

    def __init__(self, hostname: str, port: str):
        self.routes: dict = {}
        self.hostname = hostname
        self.port = port

        self.server: Optional[asyncio.AbstractServer] = None

        self.stopped = False

        self.executor = ThreadPoolExecutor(max_workers=5)

        self.logger = logging.getLogger("karp")
        self.logger.setLevel(logging.DEBUG)

        self.clients: Dict[str, Client] = {}

    async def start(self) -> None:
        """
        Creates a new KARP Server and starts it
        :return: Future
        """
        self.server = await asyncio.start_server(
            self._new_connection, host=self.hostname, port=self.port
        )
        self.logger.info(f"KARPServer started on {self.hostname}:{self.port}")
        return await self.server.serve_forever()

    async def _create_response(self, request: bytes) -> bytes:
        if not request:
            return b""
        req: Request = Request.parse(request)
        try:
            fun = self.routes[req.route]
            if asyncio.iscoroutinefunction(fun):
                response_data = await fun(req)
            else:
                response_data = await asyncio.get_event_loop().run_in_executor(
                    self.executor, self.routes[req.route], req
                )
            successful = True
        except (Exception, AssertionError) as e:
            if not req.response:
                self.logger.warning(traceback.format_exc(e))
            response_data = e.__str__()
            successful = False
        if req.response:
            res = Response.create(req.request_id, response_data, successful)
            return res.to_bytes()
        return b""

    def on_new_connection(
        self, client: Client
    ):
        """
        Called on new connection.
        :param client:
        :return:
        """
        pass

    def on_connection_lost(self, client: Client):
        """
        Called, when the connection gets lost
        :param client:
        :return:
        """
        pass

    async def _handle(self, request, writer, _id, _ip) -> None:
        try:
            r = Utils.create_interaction_object(request)
            if isinstance(r, Request):
                response: bytes = await self._create_response(
                    bytes(request, "UTF-8")
                )
                if response:
                    writer.write(response)
                    await writer.drain()
            else:
                if r.request_id in self.clients[_id].requests:
                    self.clients[_id].requests[r.request_id].complete(r)
        except InvalidResponseException:
            traceback.print_exc()

    async def _new_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        _ip = writer.transport.get_extra_info("peername")
        _id = Request.generate_id()
        self.clients[_id] = Client(reader, writer, _id)
        if asyncio.iscoroutinefunction(self.on_new_connection):
            asyncio.ensure_future(self.on_new_connection(self.clients[_id]))
        else:
            asyncio.ensure_future(
                asyncio.get_event_loop().run_in_executor(
                    self.executor, self.on_new_connection, self.clients[_id]
                )
            )
        self.logger.info(f"[+] New Connection from {_ip}")
        while not self.stopped:
            try:
                requests = await reader.read()
            except (BrokenPipeError, ConnectionResetError):
                break
            if not requests:
                break

            try:
                for request in requests.decode().split("\n"):
                    if request:
                        self.logger.debug(f"= {request} =")
                        asyncio.ensure_future(
                            self._handle(request, writer, _id, _ip)
                        )
            except UnicodeDecodeError:
                break

        if asyncio.iscoroutinefunction(self.on_connection_lost):
            asyncio.ensure_future(self.on_connection_lost(self.clients[_id]))
        else:
            asyncio.ensure_future(
                asyncio.get_event_loop().run_in_executor(
                    self.executor, self.on_connection_lost, self.clients[_id]
                )
            )
        del self.clients[_id]
        self.logger.info(f"[-] Disconnected from {_ip}")

    def add_route(self, **kwargs) -> Callable:
        """
        adds a route
        :return:
        """

        if "route" not in kwargs:
            raise InvalidRouteNameException("No route name provided.")

        def _wrapper(func: Callable) -> Callable:
            self.routes[kwargs["route"]] = func
            return func

        return _wrapper

    def __del__(self) -> None:
        self.stopped = True
        self.server.close()

    def close(self) -> None:
        """
        Closes the Server
        :return:
        """
        self.stopped = True
        self.server.close()



"""
Client
"""
import asyncio
import traceback
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Dict, Optional, Callable

import karp.utils as utils
from karp.request import Request
from karp.response import InvalidResponseException, Response
from karp.server import InvalidRouteNameException


class KARPClient(object):
    """
    KARPClient
    """

    def __init__(self, hostname: str, port: str):
        """
        KARP Client
        :param hostname: to connect to
        :param port: to connect to
        """
        self.hostname = hostname
        self.port = port

        self.reader: Optional[asyncio.StreamReader] = None
        self.writer: Optional[asyncio.StreamWriter] = None

        self.executor = ThreadPoolExecutor(max_workers=5)

        self.routes: dict = {}

        self._requests: Dict[str, utils.PendingRequest] = {}

    async def _create_response(self, request: bytes) -> bytes:
        if not request:
            return b""
        req: Request = Request.parse(request)
        try:
            func = self.routes[req.route]
            if asyncio.iscoroutinefunction(func):
                response_data = await func(req)
            else:
                response_data = await asyncio.get_event_loop().run_in_executor(
                    self.executor, self.routes[req.route], req
                )
            successful = True
        except Exception as e:
            if req.response:
                traceback.print_exc()
            response_data = e.__str__()
            successful = False
        if req.response:
            res = Response.create(req.request_id, response_data, successful)
            return res.to_bytes()
        return b""

    async def _handle(self, response) -> None:
        try:
            r = utils.Utils.create_interaction_object(response)
            if isinstance(r, Request):
                res: bytes = await self._create_response(bytes(r))
                if res:
                    self.writer.write(res)
                    await self.writer.drain()
            else:
                if r.request_id in self._requests:
                    self._requests[r.request_id].complete(r)
        except InvalidResponseException:
            traceback.print_exc()

    async def _read(self) -> None:
        while True:
            try:
                responses: bytes = await self.reader.read(-1)
            except BrokenPipeError:
                break
            if not responses:
                break

            for response in responses.decode().split("\n"):
                if response:
                    asyncio.ensure_future(self._handle(response))

    def add_route(self, **kwargs) -> Callable:
        """
        adds a route
        :return:
        """

        if "route" not in kwargs:
            raise InvalidRouteNameException("No route name provided.")

        def _wrapper(func: Callable) -> Callable:
            self.routes[kwargs["route"]] = func
            return func

        return _wrapper

    async def open(self) -> asyncio.Future:
        """
        Open the connection
        :return:
        """

        self.reader, self.writer = await asyncio.open_connection(
            self.hostname, self.port
        )
        return asyncio.ensure_future(self._read())

    async def request(
        self, route: str, request_data: str, timeout=None, response=True,
    ) -> Optional[Response]:
        """
        Request something
        :param route:
        :param request_data:
        :param timeout: Timeout, None for infinite, raises TimeoutError on failure
        :param response: wait for a response
        :return:
        """
        req = Request.create(route, request_data, response)

        self.writer.write(req.to_bytes())
        if response:
            self._requests[req.request_id] = utils.PendingRequest()

        await self.writer.drain()

        if not response:
            return

        try:
            response = await self._requests[req.request_id].process(
                timeout=timeout
            )
        except TimeoutError:
            del self._requests[req.request_id]
            raise

        del self._requests[req.request_id]
        return response



"""
Response
"""

import base64
import re


class InvalidResponseException(Exception):
    """
    Exception raised on invalid response
    """

    pass


class Response(object):
    """
    Response
    """

    RESPONSE_VALID_REGEX = re.compile(
        r"KARP_HEAD1([01])([0-9]{16})C_LEN([0-9]+)KARP_DATA([A-z0-9/+=]*)KARP_END"
    )

    def __init__(
        self, request_id: str, b64_data: str, successful: bool
    ) -> None:
        self._request_id: str = request_id
        self._b64_data: str = b64_data
        self._content_length: int = len(b64_data)
        self._text: str = base64.b64decode(b64_data.encode()).decode()

        self._successful = successful

    @classmethod
    def parse(cls, raw_bytes: bytes):
        """
        Parse raw bytes
        :param raw_bytes: raw bytes
        :return:
        """
        if not raw_bytes:
            return
        match = re.match(Response.RESPONSE_VALID_REGEX, raw_bytes.decode())
        if not match:
            raise InvalidResponseException(raw_bytes)

        self = cls(match.group(2), match.group(4), match.group(1) == "1")
        if not len(self._b64_data) == self._content_length:
            raise InvalidResponseException("Invalid Content_Length", raw_bytes)
        return self

    @classmethod
    def create(cls, request_id: str, data: str, successful: bool):
        """
        Create request
        :param request_id: request id
        :param data: data
        :param successful: check if the request was successful
        :return:
        """
        if data:
            b64_e = base64.b64encode(data.encode()).decode()
        else:
            b64_e = ""
        self = cls(request_id, b64_e, successful)
        return self

    @property
    def content_length(self) -> int:
        """
        Content_Length of the response
        :return:
        """
        return self._content_length

    @property
    def text(self) -> str:
        """
        Content of the response
        :return:
        """
        return self._text

    @property
    def content(self) -> str:
        """
        Content of the response
        :return:
        """
        return self._text

    @property
    def request_id(self) -> str:
        """
        Request Id
        :return:
        """
        return self._request_id

    def __bytes__(self) -> bytes:
        """
        Response as bytes
        :return:
        """
        return f"KARP_HEAD1{int(self.successful)}{self.request_id}C_LEN{self.content_length}KARP_DATA{self._b64_data}KARP_END\n".encode()

    def to_bytes(self) -> bytes:
        """
        Response as bytes
        :return:
        """
        return bytes(self)

    @property
    def successful(self) -> bool:
        """
        Return if the request was successful
        :return:
        """
        return self._successful

    def __str__(self) -> str:
        return str(self.__dict__)



"""
Setup
"""
from setuptools import find_packages, setup

with open("README.md", "r") as f:
    t = f.read()

setup(
    name="karp",
    version="0.0.11",
    packages=find_packages(),
    url="https://github.com/tooxo/karp",
    license="None",
    author="tooxo",
    author_email="till@s.chulte.de",
    description="KARP Communication Protocol",
    long_description=t,
    long_description_content_type="text/markdown"
)



"""
Utils
"""
import asyncio
from typing import Union, Optional
from karp.response import Response
from karp.request import Request, InvalidRequestException


class Utils:
    """
    Utils
    """

    @staticmethod
    def create_interaction_object(
        unknown_object: str
    ) -> Union[Request, Response, None]:
        """

        :return:
        """
        try:
            return Request.parse(unknown_object.encode())
        except InvalidRequestException:
            return Response.parse(unknown_object.encode())


class PendingRequest:
    """
    PendingRequest
    """

    def __init__(self) -> None:
        self.response: Optional[Response] = None
        self.event = asyncio.Event()

    async def process(self, timeout: Optional[float] = None) -> Response:
        """
        Wait for the request to be finished.
        :return:
        """
        await asyncio.wait_for(self.event.wait(), timeout)
        return self.response

    def complete(self, response: Response) -> None:
        """
        Complete the pending request.
        :param response:
        :return:
        """
        self.response = response
        self.event.set()



"""
Request
"""

import base64
import random
import re


class InvalidRequestException(Exception):
    """
    Exception raised on invalid request
    """

    pass


class Request(object):
    """
    Request
    """

    VALID_REQUEST_REGEX = re.compile(
        r"KARP_HEAD([A-z]+)0([01])([0-9]{16})C_LEN([0-9]+)KARP_DATA([A-z0-9/+=]*)KARP_END"
    )

    def __init__(
        self, route: str, request_id: str, b64_data: str, response: bool
    ) -> None:
        self._route = route.lower()
        self._request_id: str = request_id
        self._content_length: int = len(b64_data)
        self._response: bool = response
        self.b64_data: str = b64_data

        self._text: str = base64.b64decode(self.b64_data.encode()).decode()

    @classmethod
    def parse(cls, raw_data: bytes):
        """
        parse a request from raw bytes
        :param raw_data:
        :return:
        """
        match = re.match(Request.VALID_REQUEST_REGEX, raw_data.decode("UTF-8"))
        if not match:
            raise InvalidRequestException(raw_data)

        self = cls(
            match.group(1),
            match.group(3),
            match.group(5),
            match.group(2) == "1",
        )

        if not len(self.b64_data) == self._content_length:
            raise InvalidRequestException("Invalid Content_Length", raw_data)
        return self

    @staticmethod
    def generate_id() -> str:
        """
        Generate a 16 character unique id
        :return:
        """
        chars = list("0123456789")
        s = ""
        for x in range(16):
            s += random.choice(chars)
        return s

    @classmethod
    def create(cls, route: str, data: str, response: bool = True):
        """
        Create new request
        :param route: route to connect to
        :param data: data to send
        :param response: response
        :return:
        """
        b64_e = base64.b64encode(data.encode()).decode()
        self = cls(route, Request.generate_id(), b64_e, response)
        return self

    @property
    def content_length(self) -> int:
        """
        Content_Length of the response
        :return:
        """
        return self._content_length

    @property
    def text(self) -> str:
        """
        Content of the response
        :return:
        """
        return self._text

    @property
    def content(self) -> str:
        """
        Content of the response
        :return:
        """
        return self._text

    @property
    def route(self) -> str:
        """
        Route
        :return:
        """
        return self._route

    @property
    def request_id(self) -> str:
        """
        Request Id
        :return:
        """
        return self._request_id

    @property
    def response(self) -> bool:
        """
        Response
        :return:
        """
        return self._response

    def __bytes__(self) -> bytes:
        """
        Request as bytes
        :return:
        """
        return (
            f"KARP_HEAD{self.route}0{int(self._response)}{self.request_id}C_LEN{self.content_length}"
            f"KARP_DATA{self.b64_data}KARP_END\n".encode()
        )

    def to_bytes(self) -> bytes:
        """
        Request as bytes
        :return:
        """
        return bytes(self)

    def __str__(self) -> str:
        return str(self.__dict__)
